@@ -17,25 +17,48 @@ from acasclient.lsthing import (BlobValue, CodeValue, FileValue, LsThingValue,
 # "bob" user name registered
 # "PROJ-00000001" registered
 
+# Constants
+ACAS_DDICT = 'ACAS DDICT'
+PROJECT_METADATA = 'project metadata'
+PROJECT = 'project'
+PROJECT_NAME = 'project name'
+PROJECT_ALIAS = 'project alias'
+STATUS = 'status'
+PROJECT_STATUS = 'project status'
+PROCEDURE_DOCUMENT = 'procedure document'
+IS_RESTRICTED = 'is restricted'
+RESTRICTED = 'restricted'
+PDF_DOCUMENT = 'pdf document'
+START_DATE = 'start date'
+NAME_KEY = 'name'
+IS_RESTRICTED_KEY = 'is_restricted'
+STATUS_KEY = 'status'
+START_DATE_KEY = 'start_date'
+PDF_DOCUMENT_KEY = 'pdf_document'
+PROCEDURE_DOCUMENT_KEY = 'procedure_document'
+
+FWD_ITX = 'relates to'
+BACK_ITX = 'is related to'
 
 class Project(SimpleLsThing):
-    ls_type = 'project'
-    ls_kind = 'project'
-    preferred_label_kind = 'project name'
+    ls_type = PROJECT
+    ls_kind = PROJECT
+    preferred_label_kind = PROJECT_NAME
 
-    def __init__(self, name=None, alias=None, start_date=None, status=None, is_restricted=True, procedure_document=None, pdf_document=None, recorded_by=None):
-        names = {'project name': name, "project alias": alias}
+    def __init__(self, name=None, alias=None, start_date=None, status=None, is_restricted=True, procedure_document=None, pdf_document=None, recorded_by=None,
+                 ls_thing=None):
+        names = {PROJECT_NAME: name, PROJECT_ALIAS: alias}
         metadata = {
-            'project metadata': {
-                'start date': start_date,
-                'project status': CodeValue(status, "project", "status", "ACAS DDICT"),
-                'is restricted': CodeValue(str(is_restricted).lower(), "project", "restricted", "ACAS DDICT"),
-                'procedure document': BlobValue(file_path=procedure_document),
-                'pdf document': FileValue(value=pdf_document)
+            PROJECT_METADATA: {
+                START_DATE: start_date,
+                PROJECT_STATUS: CodeValue(status, PROJECT, STATUS, ACAS_DDICT),
+                IS_RESTRICTED: CodeValue(str(is_restricted).lower(), PROJECT, RESTRICTED, ACAS_DDICT),
+                PROCEDURE_DOCUMENT: BlobValue(file_path=procedure_document),
+                PDF_DOCUMENT: FileValue(value=pdf_document)
             }
         }
         super(Project, self).__init__(ls_type=self.ls_type, ls_kind=self.ls_kind, names=names, recorded_by=recorded_by,
-                                      preferred_label_kind=self.preferred_label_kind, metadata=metadata)
+                                      preferred_label_kind=self.preferred_label_kind, metadata=metadata, ls_thing=ls_thing)
 
 
 class TestLsThing(unittest.TestCase):
@@ -82,10 +105,10 @@ class TestLsThing(unittest.TestCase):
         """Test saving simple ls thing."""
         name = str(uuid.uuid4())
         meta_dict = {
-            "name": name,
-            "is_restricted": True,
-            "status": "active",
-            "start_date": time.time()
+            NAME_KEY: name,
+            IS_RESTRICTED_KEY: True,
+            STATUS_KEY: "active",
+            START_DATE_KEY: time.time()
         }
         newProject = Project(recorded_by=self.client.username, **meta_dict)
         self.assertIsNone(newProject.code_name)
@@ -108,71 +131,70 @@ class TestLsThing(unittest.TestCase):
 
         # Save with Path path
         meta_dict = {
-            "name": name,
-            "is_restricted": True,
-            "status": "active",
-            "start_date": time.time(),
-            "procedure_document": blob_test_path
+            NAME_KEY: name,
+            IS_RESTRICTED_KEY: True,
+            STATUS_KEY: "active",
+            START_DATE_KEY: time.time(),
+            PROCEDURE_DOCUMENT_KEY: blob_test_path
         }
         newProject = Project(recorded_by=self.client.username, **meta_dict)
         newProject.save(self.client)
-        self._check_blob_equal(newProject.metadata['project metadata']['procedure document'], file_name, file_bytes)
+        self._check_blob_equal(newProject.metadata[PROJECT_METADATA][PROCEDURE_DOCUMENT], file_name, file_bytes)
 
         # Save with string path
         meta_dict = {
-            "name": name,
-            "is_restricted": True,
-            "status": "active",
-            "start_date": time.time(),
-            "procedure_document": str(blob_test_path)
+            NAME_KEY: name,
+            IS_RESTRICTED_KEY: True,
+            STATUS_KEY: "active",
+            START_DATE_KEY: time.time(),
+            PROCEDURE_DOCUMENT_KEY: str(blob_test_path)
         }
         newProject = Project(recorded_by=self.client.username, **meta_dict)
         newProject.save(self.client)
-        self._check_blob_equal(newProject.metadata['project metadata']['procedure document'], file_name, file_bytes)
+        self._check_blob_equal(newProject.metadata[PROJECT_METADATA][PROCEDURE_DOCUMENT], file_name, file_bytes)
 
         # Write to a file by providing a full file path
         custom_file_name = "my.png"
-        output_file = newProject.metadata['project metadata']['procedure document'].write_to_file(full_file_path=Path(self.tempdir, custom_file_name))
+        output_file = newProject.metadata[PROJECT_METADATA][PROCEDURE_DOCUMENT].write_to_file(full_file_path=Path(self.tempdir, custom_file_name))
         self.assertTrue(output_file.exists())
         self.assertEqual(output_file.name, custom_file_name)
 
         # Write to a file by providing a full file path as a string
         custom_file_name = "my.png"
-        output_file = newProject.metadata['project metadata']['procedure document'].write_to_file(full_file_path=str(Path(self.tempdir, custom_file_name)))
+        output_file = newProject.metadata[PROJECT_METADATA][PROCEDURE_DOCUMENT].write_to_file(full_file_path=str(Path(self.tempdir, custom_file_name)))
         self.assertTrue(output_file.exists())
         self.assertEqual(output_file.name, custom_file_name)
 
         # Write to a file by providing a folder
-        output_file = newProject.metadata['project metadata']['procedure document'].write_to_file(folder_path=self.tempdir)
+        output_file = newProject.metadata[PROJECT_METADATA][PROCEDURE_DOCUMENT].write_to_file(folder_path=self.tempdir)
         self.assertTrue(output_file.exists())
         self.assertEqual(output_file.name, file_name)
 
         # Write to a file by providing a folder as a string
-        output_file = newProject.metadata['project metadata']['procedure document'].write_to_file(folder_path=str(self.tempdir))
+        output_file = newProject.metadata[PROJECT_METADATA][PROCEDURE_DOCUMENT].write_to_file(folder_path=str(self.tempdir))
         self.assertTrue(output_file.exists())
         self.assertEqual(output_file.name, file_name)
 
-
-        # Write to a file by providing a folder and custom file name
-        output_file = newProject.metadata['project metadata']['procedure document'].write_to_file(folder_path=self.tempdir, file_name=custom_file_name)
+        # Write to a file by providing a folder and custom file name 
+        output_file = newProject.metadata[PROJECT_METADATA][PROCEDURE_DOCUMENT].write_to_file(folder_path=self.tempdir, file_name=custom_file_name)
         self.assertTrue(output_file.exists())
         self.assertEqual(output_file.name, custom_file_name)
 
         # Write to a bad folder path fails gracefully
         with self.assertRaises(ValueError):
-            output_file = newProject.metadata['project metadata']['procedure document'].write_to_file(folder_path="GARBAGE")
+            output_file = newProject.metadata[PROJECT_METADATA][PROCEDURE_DOCUMENT].write_to_file(folder_path="GARBAGE")
         try:
-            output_file = newProject.metadata['project metadata']['procedure document'].write_to_file(folder_path="GARBAGE")
+            output_file = newProject.metadata[PROJECT_METADATA][PROCEDURE_DOCUMENT].write_to_file(folder_path="GARBAGE")
         except ValueError as err:
             self.assertIn("does not exist", err.args[0])
 
         # Make sure a bad path fails gracefully
         meta_dict = {
-            "name": name,
-            "is_restricted": True,
-            "status": "active",
-            "start_date": time.time(),
-            "procedure_document": "SOMEGARBAGEPATH"
+            NAME_KEY: name,
+            IS_RESTRICTED_KEY: True,
+            STATUS_KEY: "active",
+            START_DATE_KEY: time.time(),
+            PROCEDURE_DOCUMENT_KEY: "SOMEGARBAGEPATH"
         }
         with self.assertRaises(ValueError):
             newProject = Project(recorded_by=self.client.username, **meta_dict)
@@ -183,11 +205,11 @@ class TestLsThing(unittest.TestCase):
 
         # Make sure passing a directory fails gracefully
         meta_dict = {
-            "name": name,
-            "is_restricted": True,
-            "status": "active",
-            "start_date": time.time(),
-            "procedure_document": self.tempdir
+            NAME_KEY: name,
+            IS_RESTRICTED_KEY: True,
+            STATUS_KEY: "active",
+            START_DATE_KEY: time.time(),
+            PROCEDURE_DOCUMENT_KEY: self.tempdir
         }
         with self.assertRaises(ValueError):
             newProject = Project(recorded_by=self.client.username, **meta_dict)
@@ -207,24 +229,24 @@ class TestLsThing(unittest.TestCase):
 
         # Save with Path path
         meta_dict = {
-            "name": name,
-            "is_restricted": True,
-            "status": "active",
-            "start_date": time.time(),
-            "procedure_document": file_path
+            NAME_KEY: name,
+            IS_RESTRICTED_KEY: True,
+            STATUS_KEY: "active",
+            START_DATE_KEY: time.time(),
+            PROCEDURE_DOCUMENT_KEY: file_path
         }
         newProject = Project(recorded_by=self.client.username, **meta_dict)
         newProject.save(self.client)
-        self._check_blob_equal(newProject.metadata['project metadata']['procedure document'], file_name, file_bytes)
-
+        self._check_blob_equal(newProject.metadata[PROJECT_METADATA][PROCEDURE_DOCUMENT], file_name, file_bytes)
+        
         # Then update with a different file
         file_name = '1_1_Generic.xlsx'
         file_path = self._get_path(file_name)
         file_bytes = self._get_bytes(file_path)
         new_blob_val = BlobValue(file_path=file_path)
-        newProject.metadata['project metadata']['procedure document'] = new_blob_val
+        newProject.metadata[PROJECT_METADATA][PROCEDURE_DOCUMENT] = new_blob_val
         newProject.save(self.client)
-        self._check_blob_equal(newProject.metadata['project metadata']['procedure document'], file_name, file_bytes)
+        self._check_blob_equal(newProject.metadata[PROJECT_METADATA][PROCEDURE_DOCUMENT], file_name, file_bytes)
 
     def test_003_simple_ls_thing_save_with_file_value(self):
         """Test saving simple ls thing with file value."""
@@ -233,40 +255,67 @@ class TestLsThing(unittest.TestCase):
         file_test_path = Path(__file__).resolve().parent\
             .joinpath('test_acasclient', file_name)
 
-        # # Get the file bytes for testing
-        # in_file = open(file_test_path, "rb")
-        # file_bytes = in_file.read()
-        # in_file.close()
-
-        # Save with Path path
+        # Save with Path value
         meta_dict = {
-            "name": name,
-            "is_restricted": True,
-            "status": "active",
-            "start_date": time.time(),
-            "pdf_document": file_test_path
+            NAME_KEY: name,
+            IS_RESTRICTED_KEY: True,
+            STATUS_KEY: "active",
+            START_DATE_KEY: time.time(),
+            PDF_DOCUMENT_KEY: file_test_path
         }
         newProject = Project(recorded_by=self.client.username, **meta_dict)
-        newProject.upload_file_values(self.client)
         newProject.save(self.client)
         # Write file locally and compare
-        fv = newProject.metadata['project metadata']['pdf document']
+        fv = newProject.metadata[PROJECT_METADATA][PDF_DOCUMENT]
         downloaded_path = fv.download_to_disk(self.client)
         self._check_file(downloaded_path, file_name, file_test_path)
 
-        # Save with string path
+        # Save with string value
         meta_dict = {
-            "name": name,
-            "is_restricted": True,
-            "status": "active",
-            "start_date": time.time(),
-            "pdf_document": str(file_test_path)
+            NAME_KEY: name,
+            IS_RESTRICTED_KEY: True,
+            STATUS_KEY: "active",
+            START_DATE_KEY: time.time(),
+            PDF_DOCUMENT_KEY: str(file_test_path)
         }
         newProject = Project(recorded_by=self.client.username, **meta_dict)
-        newProject.upload_file_values(self.client)
         newProject.save(self.client)
         # Write file locally and compare
-        fv = newProject.metadata['project metadata']['pdf document']
+        fv = newProject.metadata[PROJECT_METADATA][PDF_DOCUMENT]
+        downloaded_path = fv.download_to_disk(self.client)
+        self._check_file(downloaded_path, file_name, file_test_path)
+
+        # Save with Path file_path
+        meta_dict = {
+            NAME_KEY: name,
+            IS_RESTRICTED_KEY: True,
+            STATUS_KEY: "active",
+            START_DATE_KEY: time.time(),
+        }
+        newProject = Project(recorded_by=self.client.username, **meta_dict)
+        newProject.metadata[PROJECT_METADATA][PDF_DOCUMENT] = FileValue(file_path=file_test_path)
+        self.assertIsNotNone(newProject.metadata[PROJECT_METADATA][PDF_DOCUMENT].value)
+        self.assertIsNotNone(newProject.metadata[PROJECT_METADATA][PDF_DOCUMENT].comments)
+        newProject.save(self.client)
+        # Write file locally and compare
+        fv = newProject.metadata[PROJECT_METADATA][PDF_DOCUMENT]
+        downloaded_path = fv.download_to_disk(self.client)
+        self._check_file(downloaded_path, file_name, file_test_path)
+
+        # Save with string file_path
+        meta_dict = {
+            NAME_KEY: name,
+            IS_RESTRICTED_KEY: True,
+            STATUS_KEY: "active",
+            START_DATE_KEY: time.time(),
+        }
+        newProject = Project(recorded_by=self.client.username, **meta_dict)
+        newProject.metadata[PROJECT_METADATA][PDF_DOCUMENT] = FileValue(file_path=str(file_test_path))
+        self.assertIsNotNone(newProject.metadata[PROJECT_METADATA][PDF_DOCUMENT].value)
+        self.assertIsNotNone(newProject.metadata[PROJECT_METADATA][PDF_DOCUMENT].comments)
+        newProject.save(self.client)
+        # Write file locally and compare
+        fv = newProject.metadata[PROJECT_METADATA][PDF_DOCUMENT]
         downloaded_path = fv.download_to_disk(self.client)
         self._check_file(downloaded_path, file_name, file_test_path)
 
@@ -303,7 +352,80 @@ class TestLsThing(unittest.TestCase):
         lskind_to_lsvalue = get_lsKind_to_lsvalue([lsthing_value])
         assert len(lskind_to_lsvalue) == 1
         assert 'bar (baz)' in lskind_to_lsvalue
-
+    
+    def test_005_create_interactions(self):
+        name = str(uuid.uuid4())
+        meta_dict = {
+            NAME_KEY: name,
+            IS_RESTRICTED_KEY: True,
+            STATUS_KEY: "active",
+            START_DATE_KEY: time.time()
+        }
+        name_2 = str(uuid.uuid4())
+        meta_dict_2 = {
+            NAME_KEY: name_2,
+            IS_RESTRICTED_KEY: True,
+            STATUS_KEY: "active",
+            START_DATE_KEY: time.time()
+        }
+        proj_1 = Project(recorded_by=self.client.username, **meta_dict)
+        proj_1.save(self.client)
+        proj_2 = Project(recorded_by=self.client.username, **meta_dict_2)
+        proj_2.save(self.client)
+        # add an interaction
+        proj_1.add_link(FWD_ITX, proj_2, recorded_by=self.client.username)
+        assert len(proj_1.links) == 1
+        # save the interaction
+        proj_1.save(self.client)
+        # Fetch project 1 and look at the interaction
+        fresh_proj_1 = Project.get_by_code(proj_1.code_name, self.client, Project.ls_type, Project.ls_kind)
+        assert len(fresh_proj_1.links) == 1
+        assert len(fresh_proj_1._ls_thing.second_ls_things) == 1
+        # check if save populated second_ls_things properly
+        # FIXME: save is not properly populating second_ls_things
+        #assert len(proj_1._ls_thing.second_ls_things) == 1
+        itx_ls_thing_ls_thing = fresh_proj_1._ls_thing.second_ls_things[0]
+        assert itx_ls_thing_ls_thing.id is not None
+        # Fetch project 2 again and look at the interaction in reverse
+        fresh_proj_2 = Project.get_by_code(proj_2.code_name, self.client, Project.ls_type, Project.ls_kind)
+        assert len(fresh_proj_2.links) == 1
+        back_itx = fresh_proj_2.links[0]
+        assert back_itx.verb == BACK_ITX
+        # Run advanced search by interaction
+        # Forward interaction query
+        second_itx_listings = [
+            {
+                "interactionType": FWD_ITX,
+                "thingType": Project.ls_type,
+                "thingKind": Project.ls_kind,
+                "operator": "=",
+                "thingCodeName": proj_2.code_name
+            }]
+        results = self.client\
+            .advanced_search_ls_things('project', 'project', None,
+                                       second_itx_listings=second_itx_listings,
+                                       codes_only=True,
+                                       max_results=1000,
+                                       combine_terms_with_and=True)
+        assert len(results) == 1
+        assert results[0] == proj_1.code_name
+        # Backwards interaction query
+        first_itx_listings = [
+            {
+                "interactionType": FWD_ITX,
+                "thingType": Project.ls_type,
+                "thingKind": Project.ls_kind,
+                "operator": "=",
+                "thingCodeName": proj_1.code_name
+            }]
+        results = self.client\
+            .advanced_search_ls_things('project', 'project', None,
+                                       first_itx_listings=first_itx_listings,
+                                       codes_only=True,
+                                       max_results=1000,
+                                       combine_terms_with_and=True)
+        assert len(results) == 1
+        assert results[0] == proj_2.code_name
 
 class TestBlobValue(unittest.TestCase):
 
